import wandb
from wandb import util
from wandb.plots.utils import (
    test_missing,
    test_types,
    encode_labels,
    deprecation_notice,
)


def explain_text(text, probas, target_names=None):
    """
    ExplainText adds support for eli5's LIME based TextExplainer.
    Arguments:
     text (str): Text to explain
     probas (black-box classification pipeline): A function which
                    takes a list of strings (documents) and returns a matrix
                    of shape (n_samples, n_classes) with probability values,
                    i.e. a row per document and a column per output label.
    Returns:
     Nothing. To see plots, go to your W&B run page.
    Example:
     wandb.log({'roc': wandb.plots.ExplainText(text, probas)})
    """
    deprecation_notice()
    eli5 = util.get_module(
        "eli5",
        required="explain_text requires the eli5 library, install with `pip install eli5`",
    )
    if test_missing(text=text, probas=probas):
        # and test_types(proba=proba)):
        wandb.termlog("Visualizing TextExplainer.")
        te = eli5.lime.TextExplainer(random_state=42)
        te.fit(text, probas)
        html = te.show_prediction(target_names=target_names)
        return wandb.Html(html.data)



from ..language import ast
from ..language.printer import print_ast
from ..type.definition import (GraphQLEnumType, GraphQLInputObjectType,
                               GraphQLList, GraphQLNonNull, GraphQLScalarType)

_empty_list = []


def is_valid_literal_value(type, value_ast):
    if isinstance(type, GraphQLNonNull):
        of_type = type.of_type
        if not value_ast:
            return [u'Expected "{}", found null.'.format(type)]

        return is_valid_literal_value(of_type, value_ast)

    if not value_ast:
        return _empty_list

    if isinstance(value_ast, ast.Variable):
        return _empty_list

    if isinstance(type, GraphQLList):
        item_type = type.of_type
        if isinstance(value_ast, ast.ListValue):
            errors = []

            for i, item_ast in enumerate(value_ast.values):
                item_errors = is_valid_literal_value(item_type, item_ast)
                for error in item_errors:
                    errors.append(u'In element #{}: {}'.format(i, error))

            return errors

        return is_valid_literal_value(item_type, value_ast)

    if isinstance(type, GraphQLInputObjectType):
        if not isinstance(value_ast, ast.ObjectValue):
            return [u'Expected "{}", found not an object.'.format(type)]

        fields = type.fields
        field_asts = value_ast.fields

        errors = []
        for provided_field_ast in field_asts:
            if provided_field_ast.name.value not in fields:
                errors.append(u'In field "{}": Unknown field.'.format(provided_field_ast.name.value))

        field_ast_map = {field_ast.name.value: field_ast for field_ast in field_asts}

        def get_field_ast_value(field_name):
            if field_name in field_ast_map:
                return field_ast_map[field_name].value

        for field_name, field in fields.items():
            subfield_errors = is_valid_literal_value(field.type, get_field_ast_value(field_name))
            errors.extend(u'In field "{}": {}'.format(field_name, e) for e in subfield_errors)

        return errors

    assert isinstance(type, (GraphQLScalarType, GraphQLEnumType)), 'Must be input type'

    parse_result = type.parse_literal(value_ast)
    if parse_result is None:
        return [u'Expected type "{}", found {}.'.format(type.name, print_ast(value_ast))]

    return _empty_list



import codecs
import os
from typing import Sequence, Type, TYPE_CHECKING, Union

from wandb import util

from ._private import MEDIA_TMP
from .base_types.media import _numpy_arrays_to_lists, Media
from .base_types.wb_value import WBValue
from .image import Image

if TYPE_CHECKING:  # pragma: no cover
    import matplotlib  # type: ignore
    import pandas as pd  # type: ignore
    import plotly  # type: ignore

    from ..wandb_artifacts import Artifact as LocalArtifact
    from ..wandb_run import Run as LocalRun

    ValToJsonType = Union[
        dict,
        "WBValue",
        Sequence["WBValue"],
        "plotly.Figure",
        "matplotlib.artist.Artist",
        "pd.DataFrame",
        object,
    ]


class Plotly(Media):
    """
    Wandb class for plotly plots.

    Arguments:
        val: matplotlib or plotly figure
    """

    _log_type = "plotly-file"

    @classmethod
    def make_plot_media(
        cls: Type["Plotly"], val: Union["plotly.Figure", "matplotlib.artist.Artist"]
    ) -> Union[Image, "Plotly"]:
        if util.is_matplotlib_typename(util.get_full_typename(val)):
            if util.matplotlib_contains_images(val):
                return Image(val)
            val = util.matplotlib_to_plotly(val)
        return cls(val)

    def __init__(self, val: Union["plotly.Figure", "matplotlib.artist.Artist"]):
        super().__init__()
        # First, check to see if the incoming `val` object is a plotfly figure
        if not util.is_plotly_figure_typename(util.get_full_typename(val)):
            # If it is not, but it is a matplotlib figure, then attempt to convert it to plotly
            if util.is_matplotlib_typename(util.get_full_typename(val)):
                if util.matplotlib_contains_images(val):
                    raise ValueError(
                        "Plotly does not currently support converting matplotlib figures containing images. \
                            You can convert the plot to a static image with `wandb.Image(plt)` "
                    )
                val = util.matplotlib_to_plotly(val)
            else:
                raise ValueError(
                    "Logged plots must be plotly figures, or matplotlib plots convertible to plotly via mpl_to_plotly"
                )

        tmp_path = os.path.join(MEDIA_TMP.name, util.generate_id() + ".plotly.json")
        val = _numpy_arrays_to_lists(val.to_plotly_json())
        with codecs.open(tmp_path, "w", encoding="utf-8") as fp:
            util.json_dump_safer(val, fp)
        self._set_file(tmp_path, is_tmp=True, extension=".plotly.json")

    @classmethod
    def get_media_subdir(cls: Type["Plotly"]) -> str:
        return os.path.join("media", "plotly")

    def to_json(self, run_or_artifact: Union["LocalRun", "LocalArtifact"]) -> dict:
        json_dict = super().to_json(run_or_artifact)
        json_dict["_type"] = self._log_type
        return json_dict



# -*- coding: utf-8 -*-
from ..error import GraphQLError
from ..language import ast
from ..pyutils.default_ordered_dict import DefaultOrderedDict
from ..type.definition import GraphQLInterfaceType, GraphQLUnionType
from ..type.directives import GraphQLIncludeDirective, GraphQLSkipDirective
from ..type.introspection import (SchemaMetaFieldDef, TypeMetaFieldDef,
                                  TypeNameMetaFieldDef)
from ..utils.type_from_ast import type_from_ast
from .values import get_argument_values, get_variable_values

Undefined = object()


class ExecutionContext(object):
    """Data that must be available at all points during query execution.

    Namely, schema of the type system that is currently executing,
    and the fragments defined in the query document"""

    __slots__ = 'schema', 'fragments', 'root_value', 'operation', 'variable_values', 'errors', 'context_value', \
                'argument_values_cache', 'executor', 'middleware', '_subfields_cache'

    def __init__(self, schema, document_ast, root_value, context_value, variable_values, operation_name, executor, middleware):
        """Constructs a ExecutionContext object from the arguments passed
        to execute, which we will pass throughout the other execution
        methods."""
        errors = []
        operation = None
        fragments = {}

        for definition in document_ast.definitions:
            if isinstance(definition, ast.OperationDefinition):
                if not operation_name and operation:
                    raise GraphQLError('Must provide operation name if query contains multiple operations.')

                if not operation_name or definition.name and definition.name.value == operation_name:
                    operation = definition

            elif isinstance(definition, ast.FragmentDefinition):
                fragments[definition.name.value] = definition

            else:
                raise GraphQLError(
                    u'GraphQL cannot execute a request containing a {}.'.format(definition.__class__.__name__),
                    definition
                )

        if not operation:
            if operation_name:
                raise GraphQLError(u'Unknown operation named "{}".'.format(operation_name))

            else:
                raise GraphQLError('Must provide an operation.')

        variable_values = get_variable_values(schema, operation.variable_definitions or [], variable_values)

        self.schema = schema
        self.fragments = fragments
        self.root_value = root_value
        self.operation = operation
        self.variable_values = variable_values
        self.errors = errors
        self.context_value = context_value
        self.argument_values_cache = {}
        self.executor = executor
        self.middleware = middleware
        self._subfields_cache = {}

    def get_field_resolver(self, field_resolver):
        if not self.middleware:
            return field_resolver
        return self.middleware.get_field_resolver(field_resolver)

    def get_argument_values(self, field_def, field_ast):
        k = field_def, field_ast
        result = self.argument_values_cache.get(k)

        if not result:
            result = self.argument_values_cache[k] = get_argument_values(field_def.args, field_ast.arguments,
                                                                         self.variable_values)

        return result

    def get_sub_fields(self, return_type, field_asts):
        k = return_type, tuple(field_asts)
        if k not in self._subfields_cache:
            subfield_asts = DefaultOrderedDict(list)
            visited_fragment_names = set()
            for field_ast in field_asts:
                selection_set = field_ast.selection_set
                if selection_set:
                    subfield_asts = collect_fields(
                        self, return_type, selection_set,
                        subfield_asts, visited_fragment_names
                    )
            self._subfields_cache[k] = subfield_asts
        return self._subfields_cache[k]


class ExecutionResult(object):
    """The result of execution. `data` is the result of executing the
    query, `errors` is null if no errors occurred, and is a
    non-empty array if an error occurred."""

    __slots__ = 'data', 'errors', 'invalid'

    def __init__(self, data=None, errors=None, invalid=False):
        self.data = data
        self.errors = errors

        if invalid:
            assert data is None

        self.invalid = invalid

    def __eq__(self, other):
        return (
            self is other or (
                isinstance(other, ExecutionResult) and
                self.data == other.data and
                self.errors == other.errors and
                self.invalid == other.invalid
            )
        )


def get_operation_root_type(schema, operation):
    op = operation.operation
    if op == 'query':
        return schema.get_query_type()

    elif op == 'mutation':
        mutation_type = schema.get_mutation_type()

        if not mutation_type:
            raise GraphQLError(
                'Schema is not configured for mutations',
                [operation]
            )

        return mutation_type

    elif op == 'subscription':
        subscription_type = schema.get_subscription_type()

        if not subscription_type:
            raise GraphQLError(
                'Schema is not configured for subscriptions',
                [operation]
            )

        return subscription_type

    raise GraphQLError(
        'Can only execute queries, mutations and subscriptions',
        [operation]
    )


def collect_fields(ctx, runtime_type, selection_set, fields, prev_fragment_names):
    """
    Given a selectionSet, adds all of the fields in that selection to
    the passed in map of fields, and returns it at the end.

    collect_fields requires the "runtime type" of an object. For a field which
    returns and Interface or Union type, the "runtime type" will be the actual
    Object type returned by that field.
    """
    for selection in selection_set.selections:
        directives = selection.directives

        if isinstance(selection, ast.Field):
            if not should_include_node(ctx, directives):
                continue

            name = get_field_entry_key(selection)
            fields[name].append(selection)

        elif isinstance(selection, ast.InlineFragment):
            if not should_include_node(
                    ctx, directives) or not does_fragment_condition_match(
                    ctx, selection, runtime_type):
                continue

            collect_fields(ctx, runtime_type, selection.selection_set, fields, prev_fragment_names)

        elif isinstance(selection, ast.FragmentSpread):
            frag_name = selection.name.value

            if frag_name in prev_fragment_names or not should_include_node(ctx, directives):
                continue

            prev_fragment_names.add(frag_name)
            fragment = ctx.fragments.get(frag_name)
            frag_directives = fragment.directives
            if not fragment or not \
                    should_include_node(ctx, frag_directives) or not \
                    does_fragment_condition_match(ctx, fragment, runtime_type):
                continue

            collect_fields(ctx, runtime_type, fragment.selection_set, fields, prev_fragment_names)

    return fields


def should_include_node(ctx, directives):
    """Determines if a field should be included based on the @include and
    @skip directives, where @skip has higher precidence than @include."""
    # TODO: Refactor based on latest code
    if directives:
        skip_ast = None

        for directive in directives:
            if directive.name.value == GraphQLSkipDirective.name:
                skip_ast = directive
                break

        if skip_ast:
            args = get_argument_values(
                GraphQLSkipDirective.args,
                skip_ast.arguments,
                ctx.variable_values,
            )
            if args.get('if') is True:
                return False

        include_ast = None

        for directive in directives:
            if directive.name.value == GraphQLIncludeDirective.name:
                include_ast = directive
                break

        if include_ast:
            args = get_argument_values(
                GraphQLIncludeDirective.args,
                include_ast.arguments,
                ctx.variable_values,
            )

            if args.get('if') is False:
                return False

    return True


def does_fragment_condition_match(ctx, fragment, type_):
    type_condition_ast = fragment.type_condition
    if not type_condition_ast:
        return True

    conditional_type = type_from_ast(ctx.schema, type_condition_ast)
    if conditional_type.is_same_type(type_):
        return True

    if isinstance(conditional_type, (GraphQLInterfaceType, GraphQLUnionType)):
        return ctx.schema.is_possible_type(conditional_type, type_)

    return False


def get_field_entry_key(node):
    """Implements the logic to compute the key of a given field's entry"""
    if node.alias:
        return node.alias.value
    return node.name.value


class ResolveInfo(object):
    __slots__ = ('field_name', 'field_asts', 'return_type', 'parent_type',
                 'schema', 'fragments', 'root_value', 'operation', 'variable_values')

    def __init__(self, field_name, field_asts, return_type, parent_type,
                 schema, fragments, root_value, operation, variable_values):
        self.field_name = field_name
        self.field_asts = field_asts
        self.return_type = return_type
        self.parent_type = parent_type
        self.schema = schema
        self.fragments = fragments
        self.root_value = root_value
        self.operation = operation
        self.variable_values = variable_values


def default_resolve_fn(source, args, context, info):
    """If a resolve function is not given, then a default resolve behavior is used which takes the property of the source object
    of the same name as the field and returns it as the result, or if it's a function, returns the result of calling that function."""
    name = info.field_name
    property = getattr(source, name, None)
    if callable(property):
        return property()
    return property


def get_field_def(schema, parent_type, field_name):
    """This method looks up the field on the given type defintion.
    It has special casing for the two introspection fields, __schema
    and __typename. __typename is special because it can always be
    queried as a field, even in situations where no other fields
    are allowed, like on a Union. __schema could get automatically
    added to the query type, but that would require mutating type
    definitions, which would cause issues."""
    if field_name == '__schema' and schema.get_query_type() == parent_type:
        return SchemaMetaFieldDef
    elif field_name == '__type' and schema.get_query_type() == parent_type:
        return TypeMetaFieldDef
    elif field_name == '__typename':
        return TypeNameMetaFieldDef
    return parent_type.fields.get(field_name)



# -*- coding: utf-8 -*-
"""
    pygments.lexers.ooc
    ~~~~~~~~~~~~~~~~~~~

    Lexers for the Ooc language.

    :copyright: Copyright 2006-2017 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, bygroups, words
from pygments.token import Text, Comment, Operator, Keyword, Name, String, \
    Number, Punctuation

__all__ = ['OocLexer']


class OocLexer(RegexLexer):
    """
    For `Ooc <http://ooc-lang.org/>`_ source code

    .. versionadded:: 1.2
    """
    name = 'Ooc'
    aliases = ['ooc']
    filenames = ['*.ooc']
    mimetypes = ['text/x-ooc']

    tokens = {
        'root': [
            (words((
                'class', 'interface', 'implement', 'abstract', 'extends', 'from',
                'this', 'super', 'new', 'const', 'final', 'static', 'import',
                'use', 'extern', 'inline', 'proto', 'break', 'continue',
                'fallthrough', 'operator', 'if', 'else', 'for', 'while', 'do',
                'switch', 'case', 'as', 'in', 'version', 'return', 'true',
                'false', 'null'), prefix=r'\b', suffix=r'\b'),
             Keyword),
            (r'include\b', Keyword, 'include'),
            (r'(cover)([ \t]+)(from)([ \t]+)(\w+[*@]?)',
             bygroups(Keyword, Text, Keyword, Text, Name.Class)),
            (r'(func)((?:[ \t]|\\\n)+)(~[a-z_]\w*)',
             bygroups(Keyword, Text, Name.Function)),
            (r'\bfunc\b', Keyword),
            # Note: %= and ^= not listed on http://ooc-lang.org/syntax
            (r'//.*', Comment),
            (r'(?s)/\*.*?\*/', Comment.Multiline),
            (r'(==?|\+=?|-[=>]?|\*=?|/=?|:=|!=?|%=?|\?|>{1,3}=?|<{1,3}=?|\.\.|'
             r'&&?|\|\|?|\^=?)', Operator),
            (r'(\.)([ \t]*)([a-z]\w*)', bygroups(Operator, Text,
                                                 Name.Function)),
            (r'[A-Z][A-Z0-9_]+', Name.Constant),
            (r'[A-Z]\w*([@*]|\[[ \t]*\])?', Name.Class),

            (r'([a-z]\w*(?:~[a-z]\w*)?)((?:[ \t]|\\\n)*)(?=\()',
             bygroups(Name.Function, Text)),
            (r'[a-z]\w*', Name.Variable),

            # : introduces types
            (r'[:(){}\[\];,]', Punctuation),

            (r'0x[0-9a-fA-F]+', Number.Hex),
            (r'0c[0-9]+', Number.Oct),
            (r'0b[01]+', Number.Bin),
            (r'[0-9_]\.[0-9_]*(?!\.)', Number.Float),
            (r'[0-9_]+', Number.Decimal),

            (r'"(?:\\.|\\[0-7]{1,3}|\\x[a-fA-F0-9]{1,2}|[^\\"])*"',
             String.Double),
            (r"'(?:\\.|\\[0-9]{1,3}|\\x[a-fA-F0-9]{1,2}|[^\\\'\n])'",
             String.Char),
            (r'@', Punctuation),  # pointer dereference
            (r'\.', Punctuation),  # imports or chain operator

            (r'\\[ \t\n]', Text),
            (r'[ \t]+', Text),
        ],
        'include': [
            (r'[\w/]+', Name),
            (r',', Punctuation),
            (r'[ \t]', Text),
            (r'[;\n]', Text, '#pop'),
        ],
    }



import wandb
from wandb import util
from wandb.plots.utils import test_missing, test_types


chart_limit = wandb.Table.MAX_ROWS


def pr_curve(y_true=None, y_probas=None, labels=None, classes_to_plot=None, title=None):
    """
    Computes the tradeoff between precision and recall for different thresholds.
    A high area under the curve represents both high recall and high precision,
    where high precision relates to a low false positive rate, and high recall
    relates to a low false negative rate. High scores for both show that the
    classifier is returning accurate results (high precision), and
    returning a majority of all positive results (high recall).
    PR curve is useful when the classes are very imbalanced.

    Arguments:
        y_true (arr): Test set labels.
        y_probas (arr): Test set predicted probabilities.
        labels (list): Named labels for target variable (y). Makes plots easier to
            read by replacing target values with corresponding index.
            For example labels = ["dog", "cat", "owl"] all 0s are
            replaced by "dog", 1s by "cat".

    Returns:
        Nothing. To see plots, go to your W&B run page then expand the 'media' tab
        under 'auto visualizations'.

    Example:
        ```
        wandb.log({"pr-curve": wandb.plot.pr_curve(y_true, y_probas, labels)})
        ```
    """
    np = util.get_module(
        "numpy",
        required="roc requires the numpy library, install with `pip install numpy`",
    )
    preprocessing = util.get_module(
        "sklearn.preprocessing",
        "roc requires the scikit preprocessing submodule, install with `pip install scikit-learn`",
    )

    metrics = util.get_module(
        "sklearn.metrics",
        "roc requires the scikit metrics submodule, install with `pip install scikit-learn`",
    )

    y_true = np.array(y_true)
    y_probas = np.array(y_probas)

    if test_missing(y_true=y_true, y_probas=y_probas) and test_types(
        y_true=y_true, y_probas=y_probas
    ):
        classes = np.unique(y_true)
        probas = y_probas

        if classes_to_plot is None:
            classes_to_plot = classes

        binarized_y_true = preprocessing.label_binarize(y_true, classes=classes)
        if len(classes) == 2:
            binarized_y_true = np.hstack((1 - binarized_y_true, binarized_y_true))

        pr_curves = {}
        indices_to_plot = np.in1d(classes, classes_to_plot)
        for i, to_plot in enumerate(indices_to_plot):
            if to_plot:
                precision, recall, _ = metrics.precision_recall_curve(
                    y_true, probas[:, i], pos_label=classes[i]
                )

                samples = 20
                sample_precision = []
                sample_recall = []
                for k in range(samples):
                    sample_precision.append(
                        precision[int(len(precision) * k / samples)]
                    )
                    sample_recall.append(recall[int(len(recall) * k / samples)])

                pr_curves[classes[i]] = (sample_precision, sample_recall)

        data = []
        count = 0
        for class_name in pr_curves.keys():
            precision, recall = pr_curves[class_name]
            for p, r in zip(precision, recall):
                # if class_names are ints and labels are set
                if labels is not None and (
                    isinstance(class_name, int) or isinstance(class_name, np.integer)
                ):
                    class_name = labels[class_name]
                # if class_names are ints and labels are not set
                # or, if class_names have something other than ints
                # (string, float, date) - user class_names
                data.append([class_name, round(p, 3), round(r, 3)])
                count += 1
                if count >= chart_limit:
                    wandb.termwarn(
                        "wandb uses only the first %d datapoints to create the plots."
                        % wandb.Table.MAX_ROWS
                    )
                    break
        table = wandb.Table(columns=["class", "precision", "recall"], data=data)
        title = title or "Precision v. Recall"
        return wandb.plot_table(
            "wandb/area-under-curve/v0",
            table,
            {"x": "recall", "y": "precision", "class": "class"},
            {"title": title},
        )



"""
wandb integration sagemaker module.
"""

from .auth import sagemaker_auth
from .config import parse_sm_config
from .resources import parse_sm_resources, parse_sm_secrets

__all__ = [
    "sagemaker_auth",
    "parse_sm_config",
    "parse_sm_secrets",
    "parse_sm_resources",
]



"""
sample.
"""

import math


class UniformSampleAccumulator:
    def __init__(self, min_samples=None):
        self._samples = min_samples or 64
        # force power of 2 samples
        self._samples = 2 ** int(math.ceil(math.log(self._samples, 2)))
        # target oversample by factor of 2
        self._samples2 = self._samples * 2
        # max size of each buffer
        self._max = self._samples2 // 2
        self._shift = 0
        self._mask = (1 << self._shift) - 1
        self._buckets = int(math.log(self._samples2, 2))
        self._buckets_bits = int(math.log(self._buckets, 2))
        self._buckets_mask = (1 << self._buckets_bits + 1) - 1
        self._buckets_index = 0
        self._bucket = []
        self._index = [0] * self._buckets
        self._count = 0
        self._log2 = [0]

        # pre-allocate buckets
        for _ in range(self._buckets):
            self._bucket.append([0] * self._max)
        # compute integer log2
        self._log2 += [int(math.log(i, 2)) for i in range(1, 2**self._buckets + 1)]

    def _show(self):
        print("=" * 20)
        for b in range(self._buckets):
            b = (b + self._buckets_index) % self._buckets
            vals = [self._bucket[b][i] for i in range(self._index[b])]
            print(f"{b}: {vals}")

    def add(self, val):
        self._count += 1
        cnt = self._count
        if cnt & self._mask:
            return
        b = cnt >> self._shift
        b = self._log2[b]  # b = int(math.log(b, 2))
        if b >= self._buckets:
            self._index[self._buckets_index] = 0
            self._buckets_index = (self._buckets_index + 1) % self._buckets
            self._shift += 1
            self._mask = (self._mask << 1) | 1
            b += self._buckets - 1
        b = (b + self._buckets_index) % self._buckets
        self._bucket[b][self._index[b]] = val
        self._index[b] += 1

    def get(self):
        full = []
        sampled = []
        # self._show()
        for b in range(self._buckets):
            max_num = 2**b
            b = (b + self._buckets_index) % self._buckets
            modb = self._index[b] // max_num
            for i in range(self._index[b]):
                if not modb or i % modb == 0:
                    sampled.append(self._bucket[b][i])
                full.append(self._bucket[b][i])
        if len(sampled) < self._samples:
            return tuple(full)
        return tuple(sampled)



"""
api.
"""

import requests
from urllib3.exceptions import InsecureRequestWarning
from wandb import env
from wandb import termwarn
from wandb import util

if env.ssl_disabled():
    # Because third party libraries may also use requests, we monkey patch it globally
    # and turn off urllib3 warnings instead printing a global warning to the user.
    termwarn(
        "Disabling SSL verification.  Connections to this server are not verified and may be insecure!"
    )
    requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
    old_merge_environment_settings = requests.Session.merge_environment_settings

    def merge_environment_settings(self, url, proxies, stream, verify, cert):
        settings = old_merge_environment_settings(
            self, url, proxies, stream, verify, cert
        )
        settings["verify"] = False
        return settings

    requests.Session.merge_environment_settings = merge_environment_settings

reset_path = util.vendor_setup()

from .internal import Api as InternalApi  # noqa
from .public import Api as PublicApi  # noqa

reset_path()

__all__ = ["InternalApi", "PublicApi"]



import logging
import os
from typing import Any, Dict, List, Optional, Tuple

from wandb.apis.internal import Api
from wandb.errors import ExecutionError, LaunchError
import yaml

from ._project_spec import create_project_from_spec, fetch_and_validate_project
from .agent import LaunchAgent
from .builder import loader as builder_loader
from .runner import loader
from .runner.abstract import AbstractRun
from .utils import (
    construct_launch_spec,
    LAUNCH_CONFIG_FILE,
    PROJECT_DOCKER_ARGS,
    PROJECT_SYNCHRONOUS,
    resolve_build_and_registry_config,
)

_logger = logging.getLogger(__name__)


def resolve_agent_config(
    api: Api,
    entity: Optional[str],
    project: Optional[str],
    max_jobs: Optional[float],
    queues: Optional[List[str]],
) -> Tuple[Dict[str, Any], Api]:
    defaults = {
        "entity": api.default_entity,
        "max_jobs": 1,
        "queues": ["default"],
        "api_key": api.api_key,
        "base_url": api.settings("base_url"),
        "registry": {},
        "build": {},
    }

    resolved_config: Dict[str, Any] = defaults
    if os.path.exists(os.path.expanduser(LAUNCH_CONFIG_FILE)):
        config = {}
        with open(os.path.expanduser(LAUNCH_CONFIG_FILE)) as f:
            try:
                config = yaml.safe_load(f)
                print(config)
            except yaml.YAMLError as e:
                raise LaunchError(f"Invalid launch agent config: {e}")
        resolved_config.update(dict(config))
    if os.environ.get("WANDB_PROJECT") is not None:
        resolved_config.update({"project": os.environ.get("WANDB_PROJECT")})
    if os.environ.get("WANDB_ENTITY") is not None:
        resolved_config.update({"entity": os.environ.get("WANDB_ENTITY")})
    if os.environ.get("WANDB_API_KEY") is not None:
        resolved_config.update({"api_key": os.environ.get("WANDB_API_KEY")})
    if os.environ.get("WANDB_LAUNCH_MAX_JOBS") is not None:
        resolved_config.update({"max_jobs": os.environ.get("WANDB_LAUNCH_MAX_JOBS")})
    if os.environ.get("WANDB_BASE_URL") is not None:
        resolved_config.update({"base_url": os.environ.get("WANDB_BASE_URL")})

    if project is not None:
        resolved_config.update({"project": project})
    if entity is not None:
        resolved_config.update({"entity": entity})
    if max_jobs is not None:
        resolved_config.update({"max_jobs": max_jobs})

    if queues is not None:
        resolved_config.update({"queues": queues})
    if (
        resolved_config["entity"] != defaults["entity"]
        or resolved_config["api_key"] != defaults["api_key"]
        or resolved_config["base_url"] != defaults["base_url"]
    ):
        settings = dict(
            api_key=resolved_config["api_key"],
            base_url=resolved_config["base_url"],
            project=resolved_config["project"],
            entity=resolved_config["entity"],
        )
        api = Api(default_settings=settings)

    return resolved_config, api


def create_and_run_agent(
    api: Api,
    config: Dict[str, Any],
) -> None:
    agent = LaunchAgent(api, config)
    agent.loop()


def _run(
    uri: Optional[str],
    name: Optional[str],
    project: Optional[str],
    entity: Optional[str],
    docker_image: Optional[str],
    entry_point: Optional[List[str]],
    version: Optional[str],
    parameters: Optional[Dict[str, Any]],
    resource: str,
    resource_args: Optional[Dict[str, Any]],
    launch_config: Optional[Dict[str, Any]],
    synchronous: Optional[bool],
    cuda: Optional[bool],
    api: Api,
) -> AbstractRun:
    """Helper that delegates to the project-running method corresponding to the passed-in backend."""
    launch_spec = construct_launch_spec(
        uri,
        api,
        name,
        project,
        entity,
        docker_image,
        resource,
        entry_point,
        version,
        parameters,
        resource_args,
        launch_config,
        cuda,
    )
    launch_project = create_project_from_spec(launch_spec, api)
    launch_project = fetch_and_validate_project(launch_project, api)

    # construct runner config.
    runner_config: Dict[str, Any] = {}
    runner_config[PROJECT_SYNCHRONOUS] = synchronous
    build_config: Optional[Dict[str, Any]] = None
    registry_config = None
    if launch_config is not None:
        given_docker_args = launch_config.get("docker", {}).get("args", {})
        build_config = launch_config.get("build")
        registry_config = launch_config.get("registry")
        runner_config[PROJECT_DOCKER_ARGS] = given_docker_args
    else:
        runner_config[PROJECT_DOCKER_ARGS] = {}

    default_launch_config = None

    if os.path.exists(os.path.expanduser(LAUNCH_CONFIG_FILE)):
        with open(os.path.expanduser(LAUNCH_CONFIG_FILE)) as f:
            default_launch_config = yaml.safe_load(f)

    build_config, registry_config = resolve_build_and_registry_config(
        default_launch_config, build_config, registry_config
    )
    builder = builder_loader.load_builder(build_config)
    backend = loader.load_backend(resource, api, runner_config)
    if backend:
        submitted_run = backend.run(launch_project, builder, registry_config)
        # this check will always pass, run is only optional in the agent case where
        # a run queue id is present on the backend config
        assert submitted_run
        return submitted_run
    else:
        raise ExecutionError(
            "Unavailable backend {}, available backends: {}".format(
                resource, ", ".join(loader.WANDB_RUNNERS)
            )
        )


def run(
    uri: Optional[str],
    api: Api,
    entry_point: Optional[List[str]] = None,
    version: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
    name: Optional[str] = None,
    resource: str = "local",
    resource_args: Optional[Dict[str, Any]] = None,
    project: Optional[str] = None,
    entity: Optional[str] = None,
    docker_image: Optional[str] = None,
    config: Optional[Dict[str, Any]] = None,
    synchronous: Optional[bool] = True,
    cuda: Optional[bool] = None,
) -> AbstractRun:
    """Run a W&B launch experiment. The project can be wandb uri or a Git URI.

    Arguments:
    uri: URI of experiment to run. A wandb run uri or a Git repository URI.
    api: An instance of a wandb Api from wandb.apis.internal.
    entry_point: Entry point to run within the project. Defaults to using the entry point used
        in the original run for wandb URIs, or main.py for git repository URIs.
    version: For Git-based projects, either a commit hash or a branch name.
    parameters: Parameters (dictionary) for the entry point command. Defaults to using the
        the parameters used to run the original run.
    name: Name run under which to launch the run.
    resource: Execution backend for the run: W&B provides built-in support for "local" backend
    resource_args: Resource related arguments for launching runs onto a remote backend.
        Will be stored on the constructed launch config under ``resource_args``.
    project: Target project to send launched run to
    entity: Target entity to send launched run to
    config: A dictionary containing the configuration for the run. May also contain
    resource specific arguments under the key "resource_args".
    synchronous: Whether to block while waiting for a run to complete. Defaults to True.
        Note that if ``synchronous`` is False and ``backend`` is "local", this
        method will return, but the current process will block when exiting until
        the local run completes. If the current process is interrupted, any
        asynchronous runs launched via this method will be terminated. If
        ``synchronous`` is True and the run fails, the current process will
        error out as well.
    cuda: Whether to build a CUDA-enabled docker image or not


    Example:
        import wandb
        project_uri = "https://github.com/wandb/examples"
        params = {"alpha": 0.5, "l1_ratio": 0.01}
        # Run W&B project and create a reproducible docker environment
        # on a local host
        api = wandb.apis.internal.Api()
        wandb.launch(project_uri, api, parameters=params)


    Returns:
        an instance of`wandb.launch.SubmittedRun` exposing information (e.g. run ID)
        about the launched run.

    Raises:
        `wandb.exceptions.ExecutionError` If a run launched in blocking mode
        is unsuccessful.
    """
    if config is None:
        config = {}

    submitted_run_obj = _run(
        uri=uri,
        name=name,
        project=project,
        entity=entity,
        docker_image=docker_image,
        entry_point=entry_point,
        version=version,
        parameters=parameters,
        resource=resource,
        resource_args=resource_args,
        launch_config=config,
        synchronous=synchronous,
        cuda=cuda,
        api=api,
    )

    return submitted_run_obj


def _wait_for(submitted_run_obj: AbstractRun) -> None:
    """Wait on the passed-in submitted run, reporting its status to the tracking server."""
    # Note: there's a small chance we fail to report the run's status to the tracking server if
    # we're interrupted before we reach the try block below
    try:
        if submitted_run_obj.wait():
            _logger.info("=== Submitted run succeeded ===")
        else:
            raise ExecutionError("Submitted run failed")
    except KeyboardInterrupt:
        _logger.error("=== Submitted run interrupted, cancelling run ===")
        submitted_run_obj.cancel()
        raise



"""
Compatibility lightgbm module.

In the future use:
    from wandb.integration.lightgbm import wandb_callback
"""

from wandb.integration.lightgbm import log_summary, wandb_callback

__all__ = ["wandb_callback", "log_summary"]



import json
import logging
import os
import subprocess
from typing import Any, Dict, List, Optional, Tuple, Union

from dockerpycreds.utils import find_executable  # type: ignore
import requests
from wandb.docker import auth
from wandb.docker import www_authenticate
from wandb.errors import DockerError


entrypoint = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "wandb-entrypoint.sh"
)
auth_config = auth.load_config()
log = logging.getLogger(__name__)


def shell(cmd: List[str]) -> Optional[str]:
    """Simple wrapper for calling docker,

    returning None on error and the output on success"""
    try:
        return (
            subprocess.check_output(["docker"] + cmd, stderr=subprocess.STDOUT)
            .decode("utf8")
            .strip()
        )
    except subprocess.CalledProcessError as e:
        print(e)
        return None


_buildx_installed = None


def is_buildx_installed() -> bool:
    """Returns `True` if docker buildx is installed and working."""
    global _buildx_installed
    if _buildx_installed is not None:
        return _buildx_installed
    if not find_executable("docker"):
        _buildx_installed = False
    else:
        help_output = shell(["buildx", "--help"])
        _buildx_installed = help_output is not None and "buildx" in help_output
    return _buildx_installed


def build(tags: List[str], file: str, context_path: str) -> str:
    command = ["buildx", "build"] if is_buildx_installed() else ["build"]
    build_tags = []
    for tag in tags:
        build_tags += ["-t", tag]
    run(
        ["docker"] + command + build_tags + ["-f", file, context_path],
        capture_stderr=False,
        capture_stdout=False,
    )
    return tags[0]


def run(
    args: List[Any],
    capture_stdout: bool = True,
    capture_stderr: bool = True,
    input: bytes = None,
    return_stderr: bool = False,
    env: Optional[Dict[str, str]] = None,
) -> Union[str, Tuple[str, str]]:
    args = [str(x) for x in args]
    subprocess_env = dict(os.environ)
    subprocess_env.update(env or {})
    if args[1] == "buildx":
        subprocess_env["DOCKER_CLI_EXPERIMENTAL"] = "enabled"
    if capture_stdout:
        stdout_dest = subprocess.PIPE
    else:
        stdout_dest = None
    if capture_stderr:
        stderr_dest = subprocess.PIPE
    else:
        stderr_dest = None

    completed_process = subprocess.run(
        args, input=input, stdout=stdout_dest, stderr=stderr_dest, env=subprocess_env
    )
    if completed_process.returncode != 0:
        raise DockerError(
            args,
            completed_process.returncode,
            completed_process.stdout,
            completed_process.stderr,
        )

    if return_stderr:
        return (
            _post_process_stream(completed_process.stdout),
            _post_process_stream(completed_process.stderr),
        )
    else:
        return _post_process_stream(completed_process.stdout)


def _post_process_stream(stream: Optional[bytes]) -> str:
    if stream is None:
        return ""
    stream = stream.decode()
    if len(stream) != 0 and stream[-1] == "\n":
        stream = stream[:-1]
    return stream


def default_image(gpu=False) -> str:
    tag = "all"
    if not gpu:
        tag += "-cpu"
    return "wandb/deepo:%s" % tag


def parse_repository_tag(repo_name: str) -> Tuple[Optional[str]]:
    parts = repo_name.rsplit("@", 1)
    if len(parts) == 2:
        return tuple(parts)
    parts = repo_name.rsplit(":", 1)
    if len(parts) == 2 and "/" not in parts[1]:
        return tuple(parts)
    return repo_name, None


def parse(image_name: str) -> Tuple[str]:
    repository, tag = parse_repository_tag(image_name)
    registry, repo_name = auth.resolve_repository_name(repository)
    if registry == "docker.io":
        registry = "index.docker.io"
    return registry, repo_name, tag or "latest"


def auth_token(registry: str, repo: str) -> Dict[str, str]:
    """Makes a request to the root of a v2 docker registry to get the auth url.

    Always returns a dictionary, if there's no token key we couldn't authenticate
    """
    # TODO: Cache tokens?
    auth_info = auth_config.resolve_authconfig(registry)
    if auth_info:
        normalized = {k.lower(): v for k, v in auth_info.items()}
        auth_info = (normalized.get("username"), normalized.get("password"))
    response = requests.get(f"https://{registry}/v2/", timeout=3)
    if response.headers.get("www-authenticate"):
        try:
            info = www_authenticate.parse(response.headers["www-authenticate"])
        except ValueError:
            info = {}
    else:
        log.error(
            "Received {} when attempting to authenticate with {}".format(
                response, registry
            )
        )
        info = {}
    if info.get("bearer"):
        res = requests.get(
            info["bearer"]["realm"]
            + "?service={}&scope=repository:{}:pull".format(
                info["bearer"]["service"], repo
            ),
            auth=auth_info,
            timeout=3,
        )
        res.raise_for_status()
        return res.json()
    return {}


def image_id_from_registry(image_name: str) -> Optional[str]:
    """Get the docker id from a public or private registry"""
    registry, repository, tag = parse(image_name)
    res = None
    try:
        token = auth_token(registry, repository).get("token")
        # dockerhub is crazy
        if registry == "index.docker.io":
            registry = "registry-1.docker.io"
        res = requests.head(
            f"https://{registry}/v2/{repository}/manifests/{tag}",
            headers={
                "Authorization": f"Bearer {token}",
                "Accept": "application/vnd.docker.distribution.manifest.v2+json",
            },
            timeout=5,
        )
        res.raise_for_status()
    except requests.RequestException:
        log.error(f"Received {res} when attempting to get digest for {image_name}")
        return None
    return "@".join([registry + "/" + repository, res.headers["Docker-Content-Digest"]])


def image_id(image_name: str) -> Optional[str]:
    """Retreve the image id from the local docker daemon or remote registry"""
    if "@sha256:" in image_name:
        return image_name
    else:
        digests = shell(["inspect", image_name, "--format", "{{json .RepoDigests}}"])
        try:
            if digests is None:
                raise ValueError()
            return json.loads(digests)[0]
        except (ValueError, IndexError):
            return image_id_from_registry(image_name)


def get_image_uid(image_name: str) -> int:
    """Retreve the image default uid through brute force"""
    return int(shell(["run", image_name, "id", "-u"]))


def push(image: str, tag: str) -> Optional[str]:
    """Push an image to a remote registry"""
    return shell(["push", f"{image}:{tag}"])


def login(username: str, password: str, registry: str) -> Optional[str]:
    """Login to a registry"""
    return shell(["login", "--username", username, "--password", password, registry])


def tag(image_name: str, tag: str) -> Optional[str]:
    """Tag an image"""
    return shell(["tag", image_name, tag])


__all__ = [
    "shell",
    "build",
    "run",
    "image_id",
    "image_id_from_registry",
    "auth_token",
    "parse",
    "parse_repository_tag",
    "default_image",
    "get_image_uid",
    "push",
    "login",
    "tag",
]



import inspect
from functools import partial
from itertools import chain

from promise import Promise

MIDDLEWARE_RESOLVER_FUNCTION = 'resolve'


class MiddlewareManager(object):

    def __init__(self, *middlewares, **kwargs):
        self.middlewares = middlewares
        self.wrap_in_promise = kwargs.get('wrap_in_promise', True)
        self._middleware_resolvers = list(get_middleware_resolvers(middlewares))
        self._cached_resolvers = {}

    def get_field_resolver(self, field_resolver):
        if field_resolver not in self._cached_resolvers:
            self._cached_resolvers[field_resolver] = middleware_chain(
                field_resolver,
                self._middleware_resolvers,
                wrap_in_promise=self.wrap_in_promise,
            )

        return self._cached_resolvers[field_resolver]


middlewares = MiddlewareManager


def get_middleware_resolvers(middlewares):
    for middleware in middlewares:
        # If the middleware is a function instead of a class
        if inspect.isfunction(middleware):
            yield middleware
        if not hasattr(middleware, MIDDLEWARE_RESOLVER_FUNCTION):
            continue
        yield getattr(middleware, MIDDLEWARE_RESOLVER_FUNCTION)


def middleware_chain(func, middlewares, wrap_in_promise):
    if not middlewares:
        return func
    if wrap_in_promise:
        middlewares = chain((func, make_it_promise), middlewares)
    else:
        middlewares = chain((func,), middlewares)
    last_func = None
    for middleware in middlewares:
        last_func = partial(middleware, last_func) if last_func else middleware

    return last_func


def make_it_promise(next, *a, **b):
    return Promise.resolve(next(*a, **b))



from typing import Any, ClassVar, Dict, List, Optional, Type, TYPE_CHECKING, Union

from pkg_resources import parse_version
from wandb import util

if TYPE_CHECKING:  # pragma: no cover
    from wandb.apis.public import Artifact as PublicArtifact

    from ...wandb_artifacts import Artifact as LocalArtifact
    from ...wandb_run import Run as LocalRun

    TypeMappingType = Dict[str, Type["WBValue"]]


def _server_accepts_client_ids() -> bool:
    # First, if we are offline, assume the backend server cannot
    # accept client IDs. Unfortunately, this is the best we can do
    # until we are sure that all local versions are > "0.11.0" max_cli_version.
    # The practical implication is that tables logged in offline mode
    # will not show up in the workspace (but will still show up in artifacts). This
    # means we never lose data, and we can still view using weave. If we decided
    # to use client ids in offline mode, then the manifests and artifact data
    # would never be resolvable and would lead to failed uploads. Our position
    # is to never lose data - and instead take the tradeoff in the UI.
    if util._is_offline():
        return False

    # If the script is online, request the max_cli_version and ensure the server
    # is of a high enough version.
    max_cli_version = util._get_max_cli_version()
    if max_cli_version is None:
        return False
    return parse_version("0.11.0") <= parse_version(max_cli_version)


class _WBValueArtifactSource:
    artifact: "PublicArtifact"
    name: Optional[str]

    def __init__(self, artifact: "PublicArtifact", name: Optional[str] = None) -> None:
        self.artifact = artifact
        self.name = name


class _WBValueArtifactTarget:
    artifact: "LocalArtifact"
    name: Optional[str]

    def __init__(self, artifact: "LocalArtifact", name: Optional[str] = None) -> None:
        self.artifact = artifact
        self.name = name


class WBValue:
    """
    Abstract parent class for things that can be logged by `wandb.log()` and
    visualized by wandb.

    The objects will be serialized as JSON and always have a _type attribute
    that indicates how to interpret the other fields.
    """

    # Class Attributes
    _type_mapping: ClassVar[Optional["TypeMappingType"]] = None
    # override _log_type to indicate the type which the subclass deserializes
    _log_type: ClassVar[Optional[str]] = None

    # Instance Attributes
    _artifact_source: Optional[_WBValueArtifactSource]
    _artifact_target: Optional[_WBValueArtifactTarget]

    def __init__(self) -> None:
        self._artifact_source = None
        self._artifact_target = None

    def to_json(self, run_or_artifact: Union["LocalRun", "LocalArtifact"]) -> dict:
        """Serializes the object into a JSON blob, using a run or artifact to store additional data.

        Args:
            run_or_artifact (wandb.Run | wandb.Artifact): the Run or Artifact for which this object should be generating
            JSON for - this is useful to to store additional data if needed.

        Returns:
            dict: JSON representation
        """
        raise NotImplementedError

    @classmethod
    def from_json(
        cls: Type["WBValue"], json_obj: dict, source_artifact: "PublicArtifact"
    ) -> "WBValue":
        """Deserialize a `json_obj` into it's class representation. If additional resources were stored in the
        `run_or_artifact` artifact during the `to_json` call, then those resources are expected to be in
        the `source_artifact`.

        Args:
            json_obj (dict): A JSON dictionary to deserialize
            source_artifact (wandb.Artifact): An artifact which will hold any additional resources which were stored
            during the `to_json` function.
        """
        raise NotImplementedError

    @classmethod
    def with_suffix(cls: Type["WBValue"], name: str, filetype: str = "json") -> str:
        """Helper function to return the name with suffix added if not already

        Args:
            name (str): the name of the file
            filetype (str, optional): the filetype to use. Defaults to "json".

        Returns:
            str: a filename which is suffixed with it's `_log_type` followed by the filetype
        """
        if cls._log_type is not None:
            suffix = cls._log_type + "." + filetype
        else:
            suffix = filetype
        if not name.endswith(suffix):
            return name + "." + suffix
        return name

    @staticmethod
    def init_from_json(
        json_obj: dict, source_artifact: "PublicArtifact"
    ) -> "Optional[WBValue]":
        """Looks through all subclasses and tries to match the json obj with the class which created it. It will then
        call that subclass' `from_json` method. Importantly, this function will set the return object's `source_artifact`
        attribute to the passed in source artifact. This is critical for artifact bookkeeping. If you choose to create
        a wandb.Value via it's `from_json` method, make sure to properly set this `artifact_source` to avoid data duplication.

        Args:
            json_obj (dict): A JSON dictionary to deserialize. It must contain a `_type` key. The value of
            this key is used to lookup the correct subclass to use.
            source_artifact (wandb.Artifact): An artifact which will hold any additional resources which were stored
            during the `to_json` function.

        Returns:
            wandb.Value: a newly created instance of a subclass of wandb.Value
        """
        class_option = WBValue.type_mapping().get(json_obj["_type"])
        if class_option is not None:
            obj = class_option.from_json(json_obj, source_artifact)
            obj._set_artifact_source(source_artifact)
            return obj

        return None

    @staticmethod
    def type_mapping() -> "TypeMappingType":
        """Returns a map from `_log_type` to subclass. Used to lookup correct types for deserialization.

        Returns:
            dict: dictionary of str:class
        """
        if WBValue._type_mapping is None:
            WBValue._type_mapping = {}
            frontier = [WBValue]
            explored = set()
            while len(frontier) > 0:
                class_option = frontier.pop()
                explored.add(class_option)
                if class_option._log_type is not None:
                    WBValue._type_mapping[class_option._log_type] = class_option
                for subclass in class_option.__subclasses__():
                    if subclass not in explored:
                        frontier.append(subclass)
        return WBValue._type_mapping

    def __eq__(self, other: object) -> bool:
        return id(self) == id(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def to_data_array(self) -> List[Any]:
        """Converts the object to a list of primitives representing the underlying data"""
        raise NotImplementedError

    def _set_artifact_source(
        self, artifact: "PublicArtifact", name: Optional[str] = None
    ) -> None:
        assert (
            self._artifact_source is None
        ), "Cannot update artifact_source. Existing source: {}/{}".format(
            self._artifact_source.artifact, self._artifact_source.name
        )
        self._artifact_source = _WBValueArtifactSource(artifact, name)

    def _set_artifact_target(
        self, artifact: "LocalArtifact", name: Optional[str] = None
    ) -> None:
        assert (
            self._artifact_target is None
        ), "Cannot update artifact_target. Existing target: {}/{}".format(
            self._artifact_target.artifact, self._artifact_target.name
        )
        self._artifact_target = _WBValueArtifactTarget(artifact, name)

    def _get_artifact_entry_ref_url(self) -> Optional[str]:
        # If the object is coming from another artifact
        if self._artifact_source and self._artifact_source.name:
            ref_entry = self._artifact_source.artifact.get_path(
                type(self).with_suffix(self._artifact_source.name)
            )
            return str(ref_entry.ref_url())
        # Else, if the object is destined for another artifact and we support client IDs
        elif (
            self._artifact_target
            and self._artifact_target.name
            and self._artifact_target.artifact._client_id is not None
            and self._artifact_target.artifact._final
            and _server_accepts_client_ids()
        ):
            return "wandb-client-artifact://{}/{}".format(
                self._artifact_target.artifact._client_id,
                type(self).with_suffix(self._artifact_target.name),
            )
        # Else if we do not support client IDs, but online, then block on upload
        # Note: this is old behavior just to stay backwards compatible
        # with older server versions. This code path should be removed
        # once those versions are no longer supported. This path uses a .wait
        # which blocks the user process on artifact upload.
        elif (
            self._artifact_target
            and self._artifact_target.name
            and self._artifact_target.artifact._logged_artifact is not None
            and not util._is_offline()
            and not _server_accepts_client_ids()
        ):
            self._artifact_target.artifact.wait()
            ref_entry = self._artifact_target.artifact.get_path(
                type(self).with_suffix(self._artifact_target.name)
            )
            return str(ref_entry.ref_url())
        return None

    def _get_artifact_entry_latest_ref_url(self) -> Optional[str]:
        if (
            self._artifact_target
            and self._artifact_target.name
            and self._artifact_target.artifact._client_id is not None
            and self._artifact_target.artifact._final
            and _server_accepts_client_ids()
        ):
            return "wandb-client-artifact://{}:latest/{}".format(
                self._artifact_target.artifact._sequence_client_id,
                type(self).with_suffix(self._artifact_target.name),
            )
        # Else if we do not support client IDs, then block on upload
        # Note: this is old behavior just to stay backwards compatible
        # with older server versions. This code path should be removed
        # once those versions are no longer supported. This path uses a .wait
        # which blocks the user process on artifact upload.
        elif (
            self._artifact_target
            and self._artifact_target.name
            and self._artifact_target.artifact._logged_artifact is not None
            and not util._is_offline()
            and not _server_accepts_client_ids()
        ):
            self._artifact_target.artifact.wait()
            ref_entry = self._artifact_target.artifact.get_path(
                type(self).with_suffix(self._artifact_target.name)
            )
            return str(ref_entry.ref_url())
        return None



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from wandb.proto import wandb_internal_pb2 as wandb_dot_proto_dot_wandb__internal__pb2
from wandb.proto import wandb_server_pb2 as wandb_dot_proto_dot_wandb__server__pb2
from wandb.proto import wandb_telemetry_pb2 as wandb_dot_proto_dot_wandb__telemetry__pb2


class InternalServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RunUpdate = channel.unary_unary(
                '/wandb_internal.InternalService/RunUpdate',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunUpdateResult.FromString,
                )
        self.Attach = channel.unary_unary(
                '/wandb_internal.InternalService/Attach',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachResponse.FromString,
                )
        self.TBSend = channel.unary_unary(
                '/wandb_internal.InternalService/TBSend',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.TBRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.TBResult.FromString,
                )
        self.RunStart = channel.unary_unary(
                '/wandb_internal.InternalService/RunStart',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartResponse.FromString,
                )
        self.GetSummary = channel.unary_unary(
                '/wandb_internal.InternalService/GetSummary',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryResponse.FromString,
                )
        self.SampledHistory = channel.unary_unary(
                '/wandb_internal.InternalService/SampledHistory',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryResponse.FromString,
                )
        self.PollExit = channel.unary_unary(
                '/wandb_internal.InternalService/PollExit',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitResponse.FromString,
                )
        self.Shutdown = channel.unary_unary(
                '/wandb_internal.InternalService/Shutdown',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownResponse.FromString,
                )
        self.RunExit = channel.unary_unary(
                '/wandb_internal.InternalService/RunExit',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitResult.FromString,
                )
        self.RunPreempting = channel.unary_unary(
                '/wandb_internal.InternalService/RunPreempting',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingResult.FromString,
                )
        self.Metric = channel.unary_unary(
                '/wandb_internal.InternalService/Metric',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricResult.FromString,
                )
        self.PartialLog = channel.unary_unary(
                '/wandb_internal.InternalService/PartialLog',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryResponse.FromString,
                )
        self.Log = channel.unary_unary(
                '/wandb_internal.InternalService/Log',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryResult.FromString,
                )
        self.Summary = channel.unary_unary(
                '/wandb_internal.InternalService/Summary',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryResult.FromString,
                )
        self.Config = channel.unary_unary(
                '/wandb_internal.InternalService/Config',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigResult.FromString,
                )
        self.Files = channel.unary_unary(
                '/wandb_internal.InternalService/Files',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesResult.FromString,
                )
        self.Output = channel.unary_unary(
                '/wandb_internal.InternalService/Output',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputResult.FromString,
                )
        self.Telemetry = channel.unary_unary(
                '/wandb_internal.InternalService/Telemetry',
                request_serializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryResult.FromString,
                )
        self.Alert = channel.unary_unary(
                '/wandb_internal.InternalService/Alert',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertResult.FromString,
                )
        self.Artifact = channel.unary_unary(
                '/wandb_internal.InternalService/Artifact',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactResult.FromString,
                )
        self.LinkArtifact = channel.unary_unary(
                '/wandb_internal.InternalService/LinkArtifact',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactRecord.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactResult.FromString,
                )
        self.ArtifactSend = channel.unary_unary(
                '/wandb_internal.InternalService/ArtifactSend',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendResponse.FromString,
                )
        self.ArtifactPoll = channel.unary_unary(
                '/wandb_internal.InternalService/ArtifactPoll',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollResponse.FromString,
                )
        self.CheckVersion = channel.unary_unary(
                '/wandb_internal.InternalService/CheckVersion',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionResponse.FromString,
                )
        self.Pause = channel.unary_unary(
                '/wandb_internal.InternalService/Pause',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseResponse.FromString,
                )
        self.Resume = channel.unary_unary(
                '/wandb_internal.InternalService/Resume',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeResponse.FromString,
                )
        self.Status = channel.unary_unary(
                '/wandb_internal.InternalService/Status',
                request_serializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusResponse.FromString,
                )
        self.ServerShutdown = channel.unary_unary(
                '/wandb_internal.InternalService/ServerShutdown',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownResponse.FromString,
                )
        self.ServerStatus = channel.unary_unary(
                '/wandb_internal.InternalService/ServerStatus',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusResponse.FromString,
                )
        self.ServerInformInit = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformInit',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitResponse.FromString,
                )
        self.ServerInformStart = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformStart',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartResponse.FromString,
                )
        self.ServerInformFinish = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformFinish',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishResponse.FromString,
                )
        self.ServerInformAttach = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformAttach',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachResponse.FromString,
                )
        self.ServerInformDetach = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformDetach',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachResponse.FromString,
                )
        self.ServerInformTeardown = channel.unary_unary(
                '/wandb_internal.InternalService/ServerInformTeardown',
                request_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownRequest.SerializeToString,
                response_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownResponse.FromString,
                )


class InternalServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RunUpdate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Attach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TBSend(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunStart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSummary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SampledHistory(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PollExit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Shutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunExit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunPreempting(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Metric(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PartialLog(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Log(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Summary(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Config(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Files(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Output(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Telemetry(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Alert(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Artifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LinkArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArtifactSend(self, request, context):
        """rpc messages for async operations: Send, Poll, Cancel, Release
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ArtifactPoll(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CheckVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Pause(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Resume(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Status(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerShutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformInit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformStart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformFinish(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformAttach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformDetach(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerInformTeardown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_InternalServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RunUpdate': grpc.unary_unary_rpc_method_handler(
                    servicer.RunUpdate,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunUpdateResult.SerializeToString,
            ),
            'Attach': grpc.unary_unary_rpc_method_handler(
                    servicer.Attach,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AttachResponse.SerializeToString,
            ),
            'TBSend': grpc.unary_unary_rpc_method_handler(
                    servicer.TBSend,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.TBRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.TBResult.SerializeToString,
            ),
            'RunStart': grpc.unary_unary_rpc_method_handler(
                    servicer.RunStart,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunStartResponse.SerializeToString,
            ),
            'GetSummary': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSummary,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryResponse.SerializeToString,
            ),
            'SampledHistory': grpc.unary_unary_rpc_method_handler(
                    servicer.SampledHistory,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryResponse.SerializeToString,
            ),
            'PollExit': grpc.unary_unary_rpc_method_handler(
                    servicer.PollExit,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PollExitResponse.SerializeToString,
            ),
            'Shutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.Shutdown,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ShutdownResponse.SerializeToString,
            ),
            'RunExit': grpc.unary_unary_rpc_method_handler(
                    servicer.RunExit,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunExitResult.SerializeToString,
            ),
            'RunPreempting': grpc.unary_unary_rpc_method_handler(
                    servicer.RunPreempting,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingResult.SerializeToString,
            ),
            'Metric': grpc.unary_unary_rpc_method_handler(
                    servicer.Metric,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.MetricResult.SerializeToString,
            ),
            'PartialLog': grpc.unary_unary_rpc_method_handler(
                    servicer.PartialLog,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryResponse.SerializeToString,
            ),
            'Log': grpc.unary_unary_rpc_method_handler(
                    servicer.Log,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.HistoryResult.SerializeToString,
            ),
            'Summary': grpc.unary_unary_rpc_method_handler(
                    servicer.Summary,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.SummaryResult.SerializeToString,
            ),
            'Config': grpc.unary_unary_rpc_method_handler(
                    servicer.Config,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ConfigResult.SerializeToString,
            ),
            'Files': grpc.unary_unary_rpc_method_handler(
                    servicer.Files,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.FilesResult.SerializeToString,
            ),
            'Output': grpc.unary_unary_rpc_method_handler(
                    servicer.Output,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.OutputResult.SerializeToString,
            ),
            'Telemetry': grpc.unary_unary_rpc_method_handler(
                    servicer.Telemetry,
                    request_deserializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryResult.SerializeToString,
            ),
            'Alert': grpc.unary_unary_rpc_method_handler(
                    servicer.Alert,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.AlertResult.SerializeToString,
            ),
            'Artifact': grpc.unary_unary_rpc_method_handler(
                    servicer.Artifact,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactResult.SerializeToString,
            ),
            'LinkArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.LinkArtifact,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactRecord.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactResult.SerializeToString,
            ),
            'ArtifactSend': grpc.unary_unary_rpc_method_handler(
                    servicer.ArtifactSend,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendResponse.SerializeToString,
            ),
            'ArtifactPoll': grpc.unary_unary_rpc_method_handler(
                    servicer.ArtifactPoll,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollResponse.SerializeToString,
            ),
            'CheckVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.CheckVersion,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionResponse.SerializeToString,
            ),
            'Pause': grpc.unary_unary_rpc_method_handler(
                    servicer.Pause,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.PauseResponse.SerializeToString,
            ),
            'Resume': grpc.unary_unary_rpc_method_handler(
                    servicer.Resume,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.ResumeResponse.SerializeToString,
            ),
            'Status': grpc.unary_unary_rpc_method_handler(
                    servicer.Status,
                    request_deserializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__internal__pb2.StatusResponse.SerializeToString,
            ),
            'ServerShutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerShutdown,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownResponse.SerializeToString,
            ),
            'ServerStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerStatus,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerStatusResponse.SerializeToString,
            ),
            'ServerInformInit': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformInit,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitResponse.SerializeToString,
            ),
            'ServerInformStart': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformStart,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartResponse.SerializeToString,
            ),
            'ServerInformFinish': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformFinish,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishResponse.SerializeToString,
            ),
            'ServerInformAttach': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformAttach,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachResponse.SerializeToString,
            ),
            'ServerInformDetach': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformDetach,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachResponse.SerializeToString,
            ),
            'ServerInformTeardown': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerInformTeardown,
                    request_deserializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownRequest.FromString,
                    response_serializer=wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'wandb_internal.InternalService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class InternalService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RunUpdate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunUpdate',
            wandb_dot_proto_dot_wandb__internal__pb2.RunRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunUpdateResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Attach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Attach',
            wandb_dot_proto_dot_wandb__internal__pb2.AttachRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.AttachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TBSend(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/TBSend',
            wandb_dot_proto_dot_wandb__internal__pb2.TBRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.TBResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunStart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunStart',
            wandb_dot_proto_dot_wandb__internal__pb2.RunStartRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunStartResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSummary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/GetSummary',
            wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.GetSummaryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SampledHistory(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/SampledHistory',
            wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.SampledHistoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PollExit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/PollExit',
            wandb_dot_proto_dot_wandb__internal__pb2.PollExitRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.PollExitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Shutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Shutdown',
            wandb_dot_proto_dot_wandb__internal__pb2.ShutdownRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ShutdownResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunExit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunExit',
            wandb_dot_proto_dot_wandb__internal__pb2.RunExitRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunExitResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunPreempting(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/RunPreempting',
            wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.RunPreemptingResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Metric(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Metric',
            wandb_dot_proto_dot_wandb__internal__pb2.MetricRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.MetricResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PartialLog(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/PartialLog',
            wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.PartialHistoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Log(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Log',
            wandb_dot_proto_dot_wandb__internal__pb2.HistoryRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.HistoryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Summary(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Summary',
            wandb_dot_proto_dot_wandb__internal__pb2.SummaryRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.SummaryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Config(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Config',
            wandb_dot_proto_dot_wandb__internal__pb2.ConfigRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ConfigResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Files(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Files',
            wandb_dot_proto_dot_wandb__internal__pb2.FilesRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.FilesResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Output(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Output',
            wandb_dot_proto_dot_wandb__internal__pb2.OutputRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.OutputResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Telemetry(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Telemetry',
            wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__telemetry__pb2.TelemetryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Alert(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Alert',
            wandb_dot_proto_dot_wandb__internal__pb2.AlertRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.AlertResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Artifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Artifact',
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LinkArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/LinkArtifact',
            wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactRecord.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.LinkArtifactResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArtifactSend(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ArtifactSend',
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactSendResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ArtifactPoll(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ArtifactPoll',
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ArtifactPollResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CheckVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/CheckVersion',
            wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.CheckVersionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Pause(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Pause',
            wandb_dot_proto_dot_wandb__internal__pb2.PauseRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.PauseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Resume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Resume',
            wandb_dot_proto_dot_wandb__internal__pb2.ResumeRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.ResumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Status(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/Status',
            wandb_dot_proto_dot_wandb__internal__pb2.StatusRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__internal__pb2.StatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerShutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerShutdown',
            wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerShutdownResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerStatus',
            wandb_dot_proto_dot_wandb__server__pb2.ServerStatusRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformInit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformInit',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformInitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformStart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformStart',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformStartResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformFinish(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformFinish',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformFinishResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformAttach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformAttach',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformAttachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformDetach(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformDetach',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformDetachResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerInformTeardown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/wandb_internal.InternalService/ServerInformTeardown',
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownRequest.SerializeToString,
            wandb_dot_proto_dot_wandb__server__pb2.ServerInformTeardownResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""Defines plots for regression models built with scikit-learn."""
from warnings import simplefilter

import numpy as np

import wandb

from wandb.sklearn import utils
from wandb.sklearn import calculate

from . import shared

# ignore all future warnings
simplefilter(action="ignore", category=FutureWarning)


def regressor(model, X_train, X_test, y_train, y_test, model_name="Regressor"):
    """Generates all sklearn regressor plots supported by W&B.

    The following plots are generated:
        learning curve, summary metrics, residuals plot, outlier candidates.

    Should only be called with a fitted regressor (otherwise an error is thrown).

    Arguments:
        model: (regressor) Takes in a fitted regressor.
        X_train: (arr) Training set features.
        y_train: (arr) Training set labels.
        X_test: (arr) Test set features.
        y_test: (arr) Test set labels.
        model_name: (str) Model name. Defaults to 'Regressor'

    Returns:
        None: To see plots, go to your W&B run page then expand the 'media' tab
            under 'auto visualizations'.

    Example:
    ```python
        wandb.sklearn.plot_regressor(reg, X_train, X_test, y_train, y_test, 'Ridge')
    ```
    """
    wandb.termlog("\nPlotting %s." % model_name)

    shared.summary_metrics(model, X_train, y_train, X_test, y_test)
    wandb.termlog("Logged summary metrics.")

    shared.learning_curve(model, X_train, y_train)
    wandb.termlog("Logged learning curve.")

    outlier_candidates(model, X_train, y_train)
    wandb.termlog("Logged outlier candidates.")

    residuals(model, X_train, y_train)
    wandb.termlog("Logged residuals.")


def outlier_candidates(regressor=None, X=None, y=None):
    """Measures a datapoint's influence on regression model via cook's distance.

    Instances with high influences could potentially be outliers.

    Should only be called with a fitted regressor (otherwise an error is thrown).

    Please note this function fits the model on the training set when called.

    Arguments:
        model: (regressor) Takes in a fitted regressor.
        X: (arr) Training set features.
        y: (arr) Training set labels.

    Returns:
        None: To see plots, go to your W&B run page then expand the 'media' tab
              under 'auto visualizations'.

    Example:
    ```python
        wandb.sklearn.plot_outlier_candidates(model, X, y)
    ```
    """
    is_missing = utils.test_missing(regressor=regressor, X=X, y=y)
    correct_types = utils.test_types(regressor=regressor, X=X, y=y)
    is_fitted = utils.test_fitted(regressor)

    if is_missing and correct_types and is_fitted:
        y = np.asarray(y)

        outliers_chart = calculate.outlier_candidates(regressor, X, y)
        wandb.log({"outlier_candidates": outliers_chart})


def residuals(regressor=None, X=None, y=None):
    """Measures and plots the regressor's predicted value against the residual.

    The marginal distribution of residuals is also calculated and plotted.

    Should only be called with a fitted regressor (otherwise an error is thrown).

    Please note this function fits variations of the model on the training set when called.

    Arguments:
        regressor: (regressor) Takes in a fitted regressor.
        X: (arr) Training set features.
        y: (arr) Training set labels.

    Returns:
        None: To see plots, go to your W&B run page then expand the 'media' tab
              under 'auto visualizations'.

    Example:
    ```python
        wandb.sklearn.plot_residuals(model, X, y)
    ```
    """
    not_missing = utils.test_missing(regressor=regressor, X=X, y=y)
    correct_types = utils.test_types(regressor=regressor, X=X, y=y)
    is_fitted = utils.test_fitted(regressor)

    if not_missing and correct_types and is_fitted:
        y = np.asarray(y)

        residuals_chart = calculate.residuals(regressor, X, y)
        wandb.log({"residuals": residuals_chart})



import logging
import shlex
from typing import Any, List, Optional

import wandb
from wandb.errors import LaunchError

from .abstract import AbstractRun, AbstractRunner
from .local import _run_entry_point
from .._project_spec import get_entry_point_command, LaunchProject
from ..builder.build import get_env_vars_dict
from ..utils import (
    parse_wandb_uri,
    PROJECT_SYNCHRONOUS,
    sanitize_wandb_api_key,
    validate_wandb_python_deps,
)


_logger = logging.getLogger(__name__)


class LocalProcessRunner(AbstractRunner):
    """Runner class, uses a project to create a LocallySubmittedRun.

    LocalProcessRunner is very similar to a LocalContainerRunner, except it does not
    run the command inside a docker container. Instead, it runs the
    command specified as a process directly on the bare metal machine.

    """

    def run(  # type: ignore
        self,
        launch_project: LaunchProject,
        *args,
        **kwargs,
    ) -> Optional[AbstractRun]:
        if args is not None:
            _logger.warning(f"LocalProcessRunner.run received unused args {args}")
        if kwargs is not None:
            _logger.warning(f"LocalProcessRunner.run received unused kwargs {kwargs}")

        synchronous: bool = self.backend_config[PROJECT_SYNCHRONOUS]
        entry_point = launch_project.get_single_entry_point()

        cmd: List[Any] = []

        if launch_project.uri is None:
            raise LaunchError("Launch LocalProcessRunner received empty project uri")
        if launch_project.project_dir is None:
            raise LaunchError("Launch LocalProcessRunner received empty project dir")

        # Check to make sure local python dependencies match run's requirement.txt
        source_entity, source_project, run_name = parse_wandb_uri(launch_project.uri)
        validate_wandb_python_deps(
            source_entity,
            source_project,
            run_name,
            self._api,
            launch_project.project_dir,
        )

        env_vars = get_env_vars_dict(launch_project, None, self._api)
        for env_key, env_value in env_vars.items():
            cmd += [f"{shlex.quote(env_key)}={shlex.quote(env_value)}"]

        if not self.ack_run_queue_item(launch_project):
            return None

        entry_cmd = get_entry_point_command(entry_point, launch_project.override_args)
        cmd += entry_cmd

        command_str = " ".join(cmd).strip()
        wandb.termlog(
            "Launching run on bare machine with command: {}".format(
                sanitize_wandb_api_key(command_str)
            )
        )
        run = _run_entry_point(command_str, launch_project.project_dir)
        if synchronous:
            run.wait()
        return run



"""Reliably launch and connect to backend server process (wandb service).

Backend server process can be connected to using tcp sockets or grpc transport.
"""

import os
import subprocess
import sys
import tempfile
import time
from typing import Any, Dict, Optional

from . import port_file
from .service_base import ServiceInterface
from .service_sock import ServiceSockInterface


class _Service:
    _grpc_port: Optional[int]
    _sock_port: Optional[int]
    _service_interface: ServiceInterface
    _internal_proc: Optional[subprocess.Popen]

    def __init__(self, _use_grpc: bool = False) -> None:
        self._use_grpc = _use_grpc
        self._stub = None
        self._grpc_port = None
        self._sock_port = None
        # current code only supports grpc or socket server implementation, in the
        # future we might be able to support both
        if _use_grpc:
            from .service_grpc import ServiceGrpcInterface

            self._service_interface = ServiceGrpcInterface()
        else:
            self._service_interface = ServiceSockInterface()

    def _wait_for_ports(self, fname: str, proc: subprocess.Popen = None) -> bool:
        time_max = time.time() + 30
        while time.time() < time_max:
            if proc and proc.poll():
                # process finished
                print("proc exited with", proc.returncode)
                return False
            if not os.path.isfile(fname):
                time.sleep(0.2)
                continue
            try:
                pf = port_file.PortFile()
                pf.read(fname)
                if not pf.is_valid:
                    time.sleep(0.2)
                    continue
                self._grpc_port = pf.grpc_port
                self._sock_port = pf.sock_port
            except Exception as e:
                print("Error:", e)
                return False
            return True
        return False

    def _launch_server(self) -> None:
        """Launch server and set ports."""

        # References for starting processes
        # - https://github.com/wandb/client/blob/archive/old-cli/wandb/__init__.py
        # - https://stackoverflow.com/questions/1196074/how-to-start-a-background-process-in-python

        kwargs: Dict[str, Any] = dict(close_fds=True, start_new_session=True)

        pid = os.getpid()

        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, f"port-{pid}.txt")

            pid_str = str(os.getpid())
            exec_cmd_list = [sys.executable, "-m"]
            # Add coverage collection if needed
            if os.environ.get("YEA_RUN_COVERAGE") and os.environ.get("COVERAGE_RCFILE"):
                exec_cmd_list += ["coverage", "run", "-m"]
            service_args = [
                "wandb",
                "service",
                "--port-filename",
                fname,
                "--pid",
                pid_str,
                "--debug",
            ]
            if self._use_grpc:
                service_args.append("--serve-grpc")
            else:
                service_args.append("--serve-sock")
            internal_proc = subprocess.Popen(
                exec_cmd_list + service_args,
                env=os.environ,
                **kwargs,
            )
            ports_found = self._wait_for_ports(fname, proc=internal_proc)
            assert ports_found
            self._internal_proc = internal_proc

    def start(self) -> None:
        self._launch_server()

    @property
    def grpc_port(self) -> Optional[int]:
        return self._grpc_port

    @property
    def sock_port(self) -> Optional[int]:
        return self._sock_port

    @property
    def service_interface(self) -> ServiceInterface:
        return self._service_interface

    def join(self) -> int:
        ret = 0
        if self._internal_proc:
            ret = self._internal_proc.wait()
        return ret



from warnings import simplefilter

import numpy as np
import sklearn

import wandb
from wandb.sklearn import utils

# ignore all future warnings
simplefilter(action="ignore", category=FutureWarning)


def summary_metrics(model=None, X=None, y=None, X_test=None, y_test=None):
    """Calculates summary metrics for both regressors and classifiers.

    Called by plot_summary_metrics to visualize metrics. Please use the function
    plot_summary_metrics() if you wish to visualize your summary metrics.
    """
    y, y_test = np.asarray(y), np.asarray(y_test)
    metrics = {}
    model_name = model.__class__.__name__

    y_pred = model.predict(X_test)

    if sklearn.base.is_classifier(model):

        accuracy_score = sklearn.metrics.accuracy_score(y_test, y_pred)
        metrics["accuracy_score"] = accuracy_score

        precision = sklearn.metrics.precision_score(y_test, y_pred, average="weighted")
        metrics["precision"] = precision

        recall = sklearn.metrics.recall_score(y_test, y_pred, average="weighted")
        metrics["recall"] = recall

        f1_score = sklearn.metrics.f1_score(y_test, y_pred, average="weighted")
        metrics["f1_score"] = f1_score

    elif sklearn.base.is_regressor(model):

        mae = sklearn.metrics.mean_absolute_error(y_test, y_pred)
        metrics["mae"] = mae

        mse = sklearn.metrics.mean_squared_error(y_test, y_pred)
        metrics["mse"] = mse

        r2_score = sklearn.metrics.r2_score(y_test, y_pred)
        metrics["r2_score"] = r2_score

    metrics = {name: utils.round_2(metric) for name, metric in metrics.items()}

    table = make_table(metrics, model_name)
    chart = wandb.visualize("wandb/metrics/v1", table)

    return chart


def make_table(metrics, model_name):
    columns = ["metric_name", "metric_value", "model_name"]
    table_content = [[name, value, model_name] for name, value in metrics.items()]

    table = wandb.Table(columns=columns, data=table_content)

    return table



"""Module to facilitate adding hooks to wandb actions

Usage:
    import trigger
    trigger.register('on_something', func)
    trigger.call('on_something', *args, **kwargs)
    trigger.unregister('on_something', func)
"""
from typing import Any, Callable

_triggers = {}


def reset():
    _triggers.clear()


def register(event: str, func: Callable):
    _triggers.setdefault(event, []).append(func)


def call(event_str: str, *args: Any, **kwargs: Any):
    for func in _triggers.get(event_str, []):
        func(*args, **kwargs)


def unregister(event: str, func: Callable):
    _triggers[event].remove(func)



import six
from wandb_graphql.language.parser import parse
from wandb_graphql.language.source import Source


def gql(request_string):
    if isinstance(request_string, six.string_types):
        source = Source(request_string, 'GraphQL request')
        return parse(source)
    else:
        raise Exception('Received incompatible request "{}".'.format(request_string))



import re
from types import TracebackType
from typing import ContextManager, Dict, List, Optional, Type
from typing import TYPE_CHECKING

import wandb
from wandb.proto.wandb_telemetry_pb2 import Imports as TelemetryImports
from wandb.proto.wandb_telemetry_pb2 import TelemetryRecord


# avoid cycle, use string type reference

if TYPE_CHECKING:
    from .. import wandb_run


_LABEL_TOKEN: str = "@wandbcode{"


class _TelemetryObject:
    _run: Optional["wandb_run.Run"]
    _obj: TelemetryRecord

    def __init__(
        self, run: "wandb_run.Run" = None, obj: TelemetryRecord = None
    ) -> None:
        self._run = run or wandb.run
        self._obj = obj or TelemetryRecord()

    def __enter__(self) -> TelemetryRecord:
        return self._obj

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        if not self._run:
            return
        self._run._telemetry_callback(self._obj)


def context(
    run: "wandb_run.Run" = None, obj: TelemetryRecord = None
) -> ContextManager[TelemetryRecord]:
    return _TelemetryObject(run=run, obj=obj)


MATCH_RE = re.compile(r"(?P<code>[a-zA-Z0-9_-]+)[,}](?P<rest>.*)")


def _parse_label_lines(lines: List[str]) -> Dict[str, str]:
    seen = False
    ret = {}
    for line in lines:
        idx = line.find(_LABEL_TOKEN)
        if idx < 0:
            # Stop parsing on first non token line after match
            if seen:
                break
            continue
        seen = True
        label_str = line[idx + len(_LABEL_TOKEN) :]

        # match identifier (first token without key=value syntax (optional)
        # Note: Parse is fairly permissive as it doesnt enforce strict syntax
        r = MATCH_RE.match(label_str)
        if r:
            ret["code"] = r.group("code").replace("-", "_")
            label_str = r.group("rest")

        # match rest of tokens on one line
        tokens = re.findall(
            r'([a-zA-Z0-9_]+)\s*=\s*("[a-zA-Z0-9_-]*"|[a-zA-Z0-9_-]*)[,}]', label_str
        )
        for k, v in tokens:
            ret[k] = v.strip('"').replace("-", "_")
    return ret


__all__ = [
    "TelemetryImports",
    "TelemetryRecord",
    "context",
]



import logging
import os
import shlex
import signal
import subprocess
import sys
from typing import Any, Dict, List, Optional

import wandb
from wandb.sdk.launch.builder.abstract import AbstractBuilder

from .abstract import AbstractRun, AbstractRunner, Status
from .._project_spec import get_entry_point_command, LaunchProject
from ..builder.build import (
    get_env_vars_dict,
    pull_docker_image,
)
from ..utils import (
    _is_wandb_dev_uri,
    _is_wandb_local_uri,
    PROJECT_DOCKER_ARGS,
    PROJECT_SYNCHRONOUS,
    sanitize_wandb_api_key,
)


_logger = logging.getLogger(__name__)


class LocalSubmittedRun(AbstractRun):
    """Instance of ``AbstractRun`` corresponding to a subprocess launched to run an entry point command locally."""

    def __init__(self, command_proc: "subprocess.Popen[bytes]") -> None:
        super().__init__()
        self.command_proc = command_proc

    @property
    def id(self) -> str:
        return str(self.command_proc.pid)

    def wait(self) -> bool:
        return self.command_proc.wait() == 0

    def cancel(self) -> None:
        # Interrupt child process if it hasn't already exited
        if self.command_proc.poll() is None:
            # Kill the the process tree rooted at the child if it's the leader of its own process
            # group, otherwise just kill the child
            try:
                if self.command_proc.pid == os.getpgid(self.command_proc.pid):
                    os.killpg(self.command_proc.pid, signal.SIGTERM)
                else:
                    self.command_proc.terminate()
            except OSError:
                # The child process may have exited before we attempted to terminate it, so we
                # ignore OSErrors raised during child process termination
                _logger.info(
                    "Failed to terminate child process (PID %s). The process may have already exited.",
                    self.command_proc.pid,
                )
            self.command_proc.wait()

    def get_status(self) -> Status:
        exit_code = self.command_proc.poll()
        if exit_code is None:
            return Status("running")
        if exit_code == 0:
            return Status("finished")
        return Status("failed")


class LocalContainerRunner(AbstractRunner):
    """Runner class, uses a project to create a LocallySubmittedRun."""

    def run(
        self,
        launch_project: LaunchProject,
        builder: AbstractBuilder,
        registry_config: Dict[str, Any],
    ) -> Optional[AbstractRun]:
        synchronous: bool = self.backend_config[PROJECT_SYNCHRONOUS]
        docker_args: Dict[str, Any] = self.backend_config[PROJECT_DOCKER_ARGS]
        if launch_project.cuda:
            docker_args["gpus"] = "all"

        if _is_wandb_local_uri(self._api.settings("base_url")):
            if sys.platform == "win32":
                docker_args["net"] = "host"
            else:
                docker_args["network"] = "host"
            if sys.platform == "linux" or sys.platform == "linux2":
                docker_args["add-host"] = "host.docker.internal:host-gateway"

        entry_point = launch_project.get_single_entry_point()
        env_vars = get_env_vars_dict(launch_project, entry_point, self._api)

        # When running against local port, need to swap to local docker host
        if (
            _is_wandb_local_uri(self._api.settings("base_url"))
            and sys.platform == "darwin"
        ):
            _, _, port = self._api.settings("base_url").split(":")
            env_vars["WANDB_BASE_URL"] = f"http://host.docker.internal:{port}"
        elif _is_wandb_dev_uri(self._api.settings("base_url")):
            env_vars["WANDB_BASE_URL"] = "http://host.docker.internal:9002"

        if launch_project.docker_image:
            # user has provided their own docker image
            image_uri = launch_project.docker_image
            pull_docker_image(image_uri)
            env_vars.pop("WANDB_RUN_ID")
            # if they've given an override to the entrypoint
            entry_cmd = get_entry_point_command(
                entry_point, launch_project.override_args
            )
            command_str = " ".join(
                get_docker_command(image_uri, env_vars, entry_cmd, docker_args)
            ).strip()
        else:
            assert entry_point is not None
            repository: Optional[str] = registry_config.get("url")
            image_uri = builder.build_image(
                launch_project,
                repository,
                entry_point,
                docker_args,
            )
            command_str = " ".join(
                get_docker_command(image_uri, env_vars, [""], docker_args)
            ).strip()

        if not self.ack_run_queue_item(launch_project):
            return None
        sanitized_cmd_str = sanitize_wandb_api_key(command_str)
        wandb.termlog(f"Launching run in docker with command: {sanitized_cmd_str}")
        run = _run_entry_point(command_str, launch_project.project_dir)
        if synchronous:
            run.wait()
        return run


def _run_entry_point(command: str, work_dir: Optional[str]) -> AbstractRun:
    """Run an entry point command in a subprocess.

    Arguments:
        command: Entry point command to run
        work_dir: Working directory in which to run the command

    Returns:
        An instance of `LocalSubmittedRun`
    """
    if work_dir is None:
        work_dir = os.getcwd()
    env = os.environ.copy()
    if os.name == "nt":
        # we are running on windows
        process = subprocess.Popen(
            ["cmd", "/c", command], close_fds=True, cwd=work_dir, env=env
        )
    else:
        process = subprocess.Popen(
            ["bash", "-c", command],
            close_fds=True,
            cwd=work_dir,
            env=env,
        )

    return LocalSubmittedRun(process)


def get_docker_command(
    image: str,
    env_vars: Dict[str, str],
    entry_cmd: List[str],
    docker_args: Dict[str, Any] = None,
) -> List[str]:
    """Constructs the docker command using the image and docker args.

    Arguments:
    image: a Docker image to be run
    env_vars: a dictionary of environment variables for the command
    entry_cmd: the entry point command to run
    docker_args: a dictionary of additional docker args for the command
    """
    docker_path = "docker"
    cmd: List[Any] = [docker_path, "run", "--rm"]

    # hacky handling of env vars, needs to be improved
    for env_key, env_value in env_vars.items():
        cmd += ["-e", f"{shlex.quote(env_key)}={shlex.quote(env_value)}"]

    if docker_args:
        for name, value in docker_args.items():
            # Passed just the name as boolean flag
            if isinstance(value, bool) and value:
                if len(name) == 1:
                    cmd += ["-" + shlex.quote(name)]
                else:
                    cmd += ["--" + shlex.quote(name)]
            else:
                # Passed name=value
                if len(name) == 1:
                    cmd += ["-" + shlex.quote(name), shlex.quote(str(value))]
                else:
                    cmd += ["--" + shlex.quote(name), shlex.quote(str(value))]

    cmd += [shlex.quote(image)]
    cmd += entry_cmd
    return cmd


def join(split_command: List[str]) -> str:
    """Return a shell-escaped string from *split_command*."""
    return " ".join(shlex.quote(arg) for arg in split_command)



import pytest
import mock

from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport


@mock.patch('gql.transport.requests.RequestsHTTPTransport.execute')
def test_retries(execute_mock):
    expected_retries = 3
    execute_mock.side_effect = Exception("fail")

    client = Client(
        retries=expected_retries,
        transport=RequestsHTTPTransport(url='http://swapi.graphene-python.org/graphql')
    )

    query = gql('''
    {
      myFavoriteFilm: film(id:"RmlsbToz") {
        id
        title
        episodeId
      }
    }
    ''')

    with pytest.raises(Exception):
        client.execute(query)

    assert execute_mock.call_count == expected_retries



import json
import re
import sys

from six import string_types

from ..language import ast
from ..type.definition import (GraphQLEnumType, GraphQLInputObjectType,
                               GraphQLList, GraphQLNonNull)
from ..type.scalars import GraphQLFloat


def ast_from_value(value, type=None):
    if isinstance(type, GraphQLNonNull):
        return ast_from_value(value, type.of_type)

    if value is None:
        return None

    if isinstance(value, list):
        item_type = type.of_type if isinstance(type, GraphQLList) else None
        return ast.ListValue([ast_from_value(item, item_type) for item in value])

    elif isinstance(type, GraphQLList):
        return ast_from_value(value, type.of_type)

    if isinstance(value, bool):
        return ast.BooleanValue(value)

    if isinstance(value, (int, float)):
        string_num = str(value)
        int_value = int(value)
        is_int_value = string_num.isdigit()

        if is_int_value or (int_value == value and value < sys.maxsize):
            if type == GraphQLFloat:
                return ast.FloatValue(str(float(value)))

            return ast.IntValue(str(int(value)))

        return ast.FloatValue(string_num)

    if isinstance(value, string_types):
        if isinstance(type, GraphQLEnumType) and re.match(r'^[_a-zA-Z][_a-zA-Z0-9]*$', value):
            return ast.EnumValue(value)

        return ast.StringValue(json.dumps(value)[1:-1])

    assert isinstance(value, dict)

    fields = []
    is_graph_ql_input_object_type = isinstance(type, GraphQLInputObjectType)

    for field_name, field_value in value.items():
        field_type = None
        if is_graph_ql_input_object_type:
            field_def = type.fields.get(field_name)
            field_type = field_def and field_def.type

        field_value = ast_from_value(field_value, field_type)
        if field_value:
            fields.append(ast.ObjectField(
                ast.Name(field_name),
                field_value
            ))

    return ast.ObjectValue(fields)



"""Base service abstract class.

Derived classes for grpc and socket service interfaces classes should implement
abstract methods.
"""

from abc import abstractmethod
from collections.abc import Iterable, Mapping
import datetime
import enum
from typing import Any, Dict, TYPE_CHECKING

from wandb.proto import wandb_server_pb2 as spb
from wandb.sdk.wandb_settings import Settings

if TYPE_CHECKING:
    from google.protobuf.internal.containers import MessageMap


def _pbmap_apply_dict(
    m: "MessageMap[str, spb.SettingsValue]", d: Dict[str, Any]
) -> None:

    for k, v in d.items():
        if isinstance(v, enum.Enum):
            continue
        sv = spb.SettingsValue()
        if v is None:
            sv.null_value = True
        elif isinstance(
            v, bool
        ):  # order matters when using `isinstance` since it looks at all the subclasses and isinstance(True, int) will be true
            sv.bool_value = v
        elif isinstance(v, int):
            sv.int_value = v
        elif isinstance(v, float):
            sv.float_value = v
        elif isinstance(v, str):
            sv.string_value = v
        elif isinstance(v, Iterable) and not isinstance(v, (str, bytes, Mapping)):
            sv.tuple_value.string_values.extend(v)
        elif isinstance(v, datetime.datetime):
            sv.timestamp_value = datetime.datetime.strftime(v, "%Y%m%d_%H%M%S")
        else:
            raise Exception("unsupported type")
        m[k].CopyFrom(sv)


class ServiceInterface:
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_transport(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _svc_inform_init(self, settings: Settings, run_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def _svc_inform_start(self, settings: Settings, run_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def _svc_inform_attach(self, attach_id: str) -> spb.ServerInformAttachResponse:
        raise NotImplementedError

    @abstractmethod
    def _svc_inform_finish(self, run_id: str = None) -> None:
        raise NotImplementedError

    @abstractmethod
    def _svc_inform_teardown(self, exit_code: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def _svc_connect(self, port: int) -> None:
        raise NotImplementedError



"""Batching file prepare requests to our API."""

import os
import queue
import shutil
import threading
from typing import Any, Callable, NamedTuple, Union

from wandb.filesync import step_upload
import wandb.util


class RequestUpload(NamedTuple):
    path: str
    save_name: str
    artifact_id: str
    copy: bool
    use_prepare_flow: bool
    save_fn: Callable[..., Any]
    digest: Any


class RequestStoreManifestFiles(NamedTuple):
    manifest: Any
    artifact_id: str
    save_fn: Callable[..., Any]


class RequestCommitArtifact(NamedTuple):
    artifact_id: str
    finalize: bool
    before_commit: Callable[..., Any]
    on_commit: Callable[..., Any]


class RequestFinish(NamedTuple):
    callback: Callable[..., Any]


Event = Union[
    RequestUpload, RequestStoreManifestFiles, RequestCommitArtifact, RequestFinish
]


class StepChecksum:
    def __init__(
        self,
        api,
        tempdir,
        request_queue: "queue.Queue[Event]",
        output_queue: "queue.Queue[step_upload.Event]",
        stats,
    ):
        self._api = api
        self._tempdir = tempdir
        self._request_queue = request_queue
        self._output_queue = output_queue
        self._stats = stats

        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True

    def _thread_body(self):
        while True:
            req = self._request_queue.get()
            if isinstance(req, RequestUpload):
                path = req.path
                if req.copy:
                    path = os.path.join(
                        self._tempdir.name,
                        f"{wandb.util.generate_id()}-{req.save_name}",
                    )
                    wandb.util.mkdir_exists_ok(os.path.dirname(path))
                    try:
                        # certain linux distros throw an exception when copying
                        # large files: https://bugs.python.org/issue43743
                        shutil.copy2(req.path, path)
                    except OSError:
                        shutil._USE_CP_SENDFILE = False
                        shutil.copy2(req.path, path)
                checksum = None
                if req.use_prepare_flow:
                    # passing a checksum through indicates that we'd like to use the
                    # "prepare" file upload flow, in which we prepare the files in
                    # the database before uploading them. This is currently only
                    # used for artifact manifests
                    checksum = wandb.util.md5_file(path)
                self._stats.init_file(req.save_name, os.path.getsize(path))
                self._output_queue.put(
                    step_upload.RequestUpload(
                        path,
                        req.save_name,
                        req.artifact_id,
                        checksum,
                        req.copy,
                        req.save_fn,
                        req.digest,
                    )
                )
            elif isinstance(req, RequestStoreManifestFiles):
                for entry in req.manifest.entries.values():
                    if entry.local_path:
                        # This stupid thing is needed so the closure works correctly.
                        def make_save_fn_with_entry(save_fn, entry):
                            return lambda progress_callback: save_fn(
                                entry, progress_callback
                            )

                        self._stats.init_file(
                            entry.local_path, entry.size, is_artifact_file=True
                        )
                        self._output_queue.put(
                            step_upload.RequestUpload(
                                entry.local_path,
                                entry.path,
                                req.artifact_id,
                                entry.digest,
                                False,
                                make_save_fn_with_entry(req.save_fn, entry),
                                entry.digest,
                            )
                        )
            elif isinstance(req, RequestCommitArtifact):
                self._output_queue.put(
                    step_upload.RequestCommitArtifact(
                        req.artifact_id, req.finalize, req.before_commit, req.on_commit
                    )
                )
            elif isinstance(req, RequestFinish):
                break
            else:
                raise Exception("internal error")

        self._output_queue.put(step_upload.RequestFinish(req.callback))

    def start(self):
        self._thread.start()

    def is_alive(self):
        return self._thread.is_alive()

    def finish(self):
        self._request_queue.put(RequestFinish(None))



"""Manage wandb processes.

Create a grpc manager channel.
"""

import atexit
import os
from typing import Any, Callable, Dict, Optional, TYPE_CHECKING

from wandb import env, trigger
from wandb.sdk.lib.exit_hooks import ExitHooks
from wandb.sdk.lib.proto_util import settings_dict_from_pbmap


if TYPE_CHECKING:
    from wandb.sdk.service import service
    from wandb.sdk.service.service_base import ServiceInterface
    from wandb.sdk.wandb_settings import Settings


class _ManagerToken:
    _version = "2"
    _supported_transports = {"grpc", "tcp"}
    _token_str: str
    _pid: int
    _transport: str
    _host: str
    _port: int

    def __init__(self, token: str) -> None:
        self._token_str = token
        self._parse()

    @classmethod
    def from_environment(cls) -> Optional["_ManagerToken"]:
        token = os.environ.get(env.SERVICE)
        if not token:
            return None
        return cls(token=token)

    @classmethod
    def from_params(cls, transport: str, host: str, port: int) -> "_ManagerToken":
        version = cls._version
        pid = os.getpid()
        token = "-".join([version, str(pid), transport, host, str(port)])
        return cls(token=token)

    def set_environment(self) -> None:
        os.environ[env.SERVICE] = self._token_str

    def _parse(self) -> None:
        assert self._token_str
        parts = self._token_str.split("-")
        assert len(parts) == 5, f"token must have 5 parts: {parts}"
        # TODO: make more robust?
        version, pid_str, transport, host, port_str = parts
        assert version == self._version
        assert transport in self._supported_transports
        self._pid = int(pid_str)
        self._transport = transport
        self._host = host
        self._port = int(port_str)

    def reset_environment(self) -> None:
        os.environ.pop(env.SERVICE, None)

    @property
    def token(self) -> str:
        return self._token_str

    @property
    def pid(self) -> int:
        return self._pid

    @property
    def transport(self) -> str:
        return self._transport

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port


class _Manager:
    _token: _ManagerToken
    _atexit_lambda: Optional[Callable[[], None]]
    _hooks: Optional[ExitHooks]
    _settings: "Settings"

    def __init__(self, settings: "Settings", _use_grpc: bool = False) -> None:
        # TODO: warn if user doesnt have grpc installed
        from wandb.sdk.service import service

        self._settings = settings
        self._atexit_lambda = None
        self._hooks = None

        self._service = service._Service(_use_grpc=_use_grpc)

        token = _ManagerToken.from_environment()
        if not token:
            self._service.start()
            host = "localhost"
            if _use_grpc:
                transport = "grpc"
                port = self._service.grpc_port
            else:
                transport = "tcp"
                port = self._service.sock_port
            assert port
            token = _ManagerToken.from_params(transport=transport, host=host, port=port)
            token.set_environment()
            self._atexit_setup()

        self._token = token

        port = self._token.port
        svc_iface = self._get_service_interface()
        svc_iface._svc_connect(port=port)

    def _atexit_setup(self) -> None:
        self._atexit_lambda = lambda: self._atexit_teardown()

        self._hooks = ExitHooks()
        self._hooks.hook()
        atexit.register(self._atexit_lambda)

    def _atexit_teardown(self) -> None:
        trigger.call("on_finished")
        exit_code = self._hooks.exit_code if self._hooks else 0
        self._teardown(exit_code)

    def _teardown(self, exit_code: int) -> None:
        if self._atexit_lambda:
            atexit.unregister(self._atexit_lambda)
            self._atexit_lambda = None
        self._inform_teardown(exit_code)
        result = self._service.join()
        if result and not self._settings._jupyter:
            os._exit(result)
        self._token.reset_environment()

    def _get_service(self) -> "service._Service":
        return self._service

    def _get_service_interface(self) -> "ServiceInterface":
        assert self._service
        svc_iface = self._service.service_interface
        assert svc_iface
        return svc_iface

    def _inform_init(self, settings: "Settings", run_id: str) -> None:
        svc_iface = self._get_service_interface()
        svc_iface._svc_inform_init(settings=settings, run_id=run_id)

    def _inform_start(self, settings: "Settings", run_id: str) -> None:
        svc_iface = self._get_service_interface()
        svc_iface._svc_inform_start(settings=settings, run_id=run_id)

    def _inform_attach(self, attach_id: str) -> Dict[str, Any]:
        svc_iface = self._get_service_interface()
        response = svc_iface._svc_inform_attach(attach_id=attach_id)
        return settings_dict_from_pbmap(response._settings_map)

    def _inform_finish(self, run_id: str = None) -> None:
        svc_iface = self._get_service_interface()
        svc_iface._svc_inform_finish(run_id=run_id)

    def _inform_teardown(self, exit_code: int) -> None:
        svc_iface = self._get_service_interface()
        svc_iface._svc_inform_teardown(exit_code)



# -*- coding: utf-8 -*-
"""
Terminology

"Definitions" are the generic name for top-level statements in the document.
Examples of this include:
1) Operations (such as a query)
2) Fragments

"Operations" are a generic name for requests in the document.
Examples of this include:
1) query,
2) mutation

"Selections" are the statements that can appear legally and at
single level of the query. These include:
1) field references e.g "a"
2) fragment "spreads" e.g. "...c"
3) inline fragment "spreads" e.g. "...on Type { a }"
"""
from .executor import execute
from .base import ExecutionResult
from .middleware import middlewares, MiddlewareManager


__all__ = ['execute', 'ExecutionResult', 'MiddlewareManager', 'middlewares']



import logging
from typing import Any, Dict, List

from wandb.apis.internal import Api
from wandb.errors import LaunchError

from .abstract import AbstractRunner


__logger__ = logging.getLogger(__name__)


# Statically register backend defined in wandb
WANDB_RUNNERS: List[str] = [
    "local",
    "local-container",
    "bare",
    "local-process",
    "gcp-vertex",
    "sagemaker",
    "kubernetes",
]


def load_backend(
    backend_name: str, api: Api, backend_config: Dict[str, Any]
) -> AbstractRunner:
    # Static backends
    if backend_name in ["local", "local-container"]:
        from .local import LocalContainerRunner

        return LocalContainerRunner(api, backend_config)
    elif backend_name in ["bare", "local-process"]:
        from .bare import LocalProcessRunner

        return LocalProcessRunner(api, backend_config)
    elif backend_name == "gcp-vertex":
        from .gcp_vertex import VertexRunner

        return VertexRunner(api, backend_config)
    elif backend_name == "sagemaker":
        from .aws import AWSSagemakerRunner

        return AWSSagemakerRunner(api, backend_config)
    elif backend_name == "kubernetes":
        from .kubernetes import KubernetesRunner

        return KubernetesRunner(api, backend_config)
    raise LaunchError(
        "Resource name not among available resources. Available resources: {} ".format(
            ",".join(WANDB_RUNNERS)
        )
    )



import wandb.data_types as data_types


def get_types():
    classes = map(data_types.__dict__.get, data_types.__all__)
    types = []
    for cls in classes:
        if hasattr(cls, "_log_type") and cls._log_type is not None:
            types.append(cls._log_type)
    # add table-file type because this is a special case
    # that does not have a matching _log_type for artifacts
    # and files
    types.append("table-file")
    return types


WANDB_TYPES = get_types()


def metric_is_wandb_dict(metric):
    return "_type" in list(metric.keys()) and metric["_type"] in WANDB_TYPES
